"""Tests for mdpo text utilities."""

import pytest

from mdpo.text import max_char_in_a_row, min_not_max_chars_in_a_row


@pytest.mark.parametrize(
    ('char', 'text', 'expected_result'), (
        ('`', 'hello `` ``` a `` `', 3),
        ('`', 'hello `` ``` a ```` `', 4),
        ('`', 'hello `` ``` a `` `````', 5),

        # does not work for multiple characters
        ('``', 'hello `` ``` a `` `````', 0),
    ),
)
def test_max_char_in_a_row(char, text, expected_result):
    assert max_char_in_a_row(char, text) == expected_result


@pytest.mark.parametrize(
    ('char', 'text', 'expected_result'), (
        ('`', 'hello `` ``` a `` `', 4),
        ('`', 'hello `` ``` a ```` `', 5),
        ('`', 'hello `` ``` a `` `````', 1),
        ('`', 'hello ` `` ``` a `` `````', 4),

        # as default, 1
        ('`', 'hello', 1),

        # does not work for multiple characters (returns default)
        ('``', 'hello `` ``` a `` `````', 1),
    ),
)
def test_min_not_max_chars_in_a_row(char, text, expected_result):
    assert min_not_max_chars_in_a_row(char, text) == expected_result



"""mdpo command related utilities."""

import re


COMMAND_SEARCH_RE = re.compile(
    r'<\!\-\-\s{0,1}mdpo\-([a-z\-]+)\s{0,1}([\w\s]+)?\-\->',
)


def parse_mdpo_html_command(value):
    """Parses a mdpo HTML command inside a string value.

    This function is used by md2po implementation to discover the HTML commands
    used by mdpo to customize the extraction.

    Args:
        value (str): Text where will a command will be searched.

    Returns:
        tuple: Namename of the command (not including the ``"mdpo"`` prefix)
        and its value.
    """
    command_search = COMMAND_SEARCH_RE.search(value)
    if command_search:
        return command_search.groups()
    return (None, None)



import pytest

from mdpo.po import po_escaped_string


@pytest.mark.parametrize(
    ('string', 'escaped'), (
        ('**', '\\*'),
        ('*', '\\*'),
        ('`[', '\\`'),
        ('jfsakjbfjksafs', '\\j'),
        ('\\n', '\\\\'),
        ('\n', '\\\n'),
    ),
)
def test_po_escaped_string(string, escaped):
    assert po_escaped_string(string) == escaped



"""Markdown files translator using pofiles as reference."""

import glob
import re
import textwrap

import md4c
import polib

from mdpo.command import parse_mdpo_html_command
from mdpo.io import filter_paths, to_file_content_if_is_file
from mdpo.md import (
    escape_links_titles,
    fixwrap_codespans,
    inline_untexted_links,
)
from mdpo.md4c import DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS
from mdpo.po import po_escaped_string
from mdpo.polib import *  # noqa
from mdpo.text import min_not_max_chars_in_a_row


class Po2Md:
    def __init__(self, pofiles, ignore=[], po_encoding=None, **kwargs):
        self.pofiles = [
            polib.pofile(pofilepath, encoding=po_encoding) for pofilepath in
            filter_paths(glob.glob(pofiles), ignore_paths=ignore)
        ]

        self.extensions = kwargs.get(
            'extensions',
            DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS,
        )

        self._current_msgid = ''
        self._current_msgctxt = None
        self._current_line = ''
        self._outputlines = []

        self._disable_next_line = False
        self._disable = False
        self._enable_next_line = False
        self.disabled_entries = []

        self.translations = None
        self.translations_with_msgctxt = None

        self.bold_start_string = kwargs.get('bold_start_string', '**')
        self.bold_start_string_escaped = po_escaped_string(
            self.bold_start_string,
        )

        self.bold_end_string = kwargs.get('bold_end_string', '**')
        self.bold_end_string_escaped = po_escaped_string(
            self.bold_end_string,
        )

        self.italic_start_string = kwargs.get('italic_start_string', '*')
        self.italic_start_string_escaped = po_escaped_string(
            self.italic_start_string,
        )

        self.italic_end_string = kwargs.get('italic_end_string', '*')
        self.italic_end_string_escaped = po_escaped_string(
            self.italic_end_string,
        )

        self._bold_italic_context = False

        self.code_start_string = kwargs.get('code_start_string', '`')[0]
        self.code_start_string_escaped = po_escaped_string(
            self.code_start_string,
        )

        self.code_end_string = kwargs.get('code_end_string', '`')[0]
        self.code_end_string_escaped = po_escaped_string(
            self.code_end_string,
        )

        self.link_start_string = kwargs.get('link_start_string', '[')
        self.link_end_string = kwargs.get('link_end_string', ']')

        self._enterspan_replacer = {
            md4c.SpanType.STRONG: self.bold_start_string,
            md4c.SpanType.EM: self.italic_start_string,
            md4c.SpanType.CODE: self.code_start_string,
            md4c.SpanType.A: self.link_start_string,
        }

        self._leavespan_replacer = {
            md4c.SpanType.STRONG: self.bold_end_string,
            md4c.SpanType.EM: self.italic_end_string,
            md4c.SpanType.CODE: self.code_end_string,
            md4c.SpanType.A: self.link_end_string,
        }

        if 'wikilinks' in self.extensions:
            self.wikilink_start_string = kwargs.get('link_end_string', '[[')
            self.wikilink_end_string = kwargs.get('link_end_string', ']]')

            self._enterspan_replacer[md4c.SpanType.WIKILINK] = \
                self.wikilink_start_string
            self._leavespan_replacer[md4c.SpanType.WIKILINK] = \
                self.wikilink_end_string

        self._inside_htmlblock = False
        self._inside_codeblock = False
        self._inside_indented_codeblock = False

        self._inside_pblock = False
        self._inside_liblock = False

        self._inside_codespan = False
        self._codespan_start_index = None
        self._codespan_backticks = None
        self._codespan_inside_current_msgid = False

        self._inside_quoteblock = False

        self._current_aspan_href = None
        self._current_imgspan = {}

        # current table head alignments
        self._current_thead_aligns = []

        # if title are found in images of links for the current msgid,
        # we need to escape them after translate it because ``polib.unescape``
        # removes the scapes
        self._aimg_title_inside_current_msgid = False

        # current UL marks by nesting levels
        self._ul_marks = []

        # [numerical iterm order, current delimitier] for OL blocks
        self._current_ol_delimiters = []

        self._current_wikilink_target = None

    def _process_command(self, text):
        command, comment = parse_mdpo_html_command(text)
        if command is None:
            return

        if command == 'disable-next-line':
            self._disable_next_line = True
        elif command == 'disable':
            self._disable = True
        elif command == 'enable':
            self._disable = False
        elif command == 'enable-next-line':
            self._enable_next_line = True
        elif command == 'context' and comment:
            self._current_msgctxt = comment.strip(' ')

    def _escape_translation(self, text):
        # escape '"' characters inside links and image titles
        if self._aimg_title_inside_current_msgid:
            text = escape_links_titles(
                text, link_start_string=self.link_start_string,
                link_end_string=self.link_end_string,
            )
        # - `[self-referenced-link]` -> <self-referenced-link>
        return inline_untexted_links(
            text,
            link_start_string=self.link_start_string,
            link_end_string=self.link_end_string,
        )

    def _translate_msgid(self, msgid, msgctxt):
        try:
            if msgctxt:
                response = self.translations_with_msgctxt[msgctxt][msgid]
            else:
                response = self.translations[msgid]
        except KeyError:
            response = msgid
        return response or msgid

    def _save_current_msgid(self):
        if (not self._disable and not self._disable_next_line) or \
                self._enable_next_line:
            translation = self._translate_msgid(
                self._current_msgid,
                self._current_msgctxt,
            )
        else:
            translation = self._current_msgid
            self.disabled_entries.append(
                polib.POEntry(
                    msgid=translation,
                    msgstr='',
                    msgctxt=self._current_msgctxt,
                ),
            )

        if not self._inside_codeblock:
            translation = self._escape_translation(translation)

        elif self._inside_indented_codeblock:
            # add 4 spaces before each line including next indented block code
            translation = '    %s' % re.sub('\n', '\n    ', translation)
        if self._inside_liblock:
            translation = '\n'.join(textwrap.wrap(translation, width=79))
        if self._inside_pblock:
            # wrap paragraphs fitting with markdownlint
            if self._codespan_inside_current_msgid:
                # fix codespans wrapping
                lines = fixwrap_codespans(translation.split('\n'))
            else:
                lines = textwrap.wrap(translation, width=80)
            for line in lines:
                self._current_line += '%s\n' % polib.unescape(line)
        else:
            self._current_line += polib.unescape(translation)
        self._current_msgid = ''
        self._current_msgctxt = None

        self._disable_next_line = False
        self._enable_next_line = False

        self._codespan_inside_current_msgid = False
        self._aimg_title_inside_current_msgid = False

    def _save_current_line(self):
        self._outputlines.append(self._current_line.rstrip(' '))
        self._current_line = ''

    def enter_block(self, block, details):
        # print('ENTER BLOCK', block.name, details)

        if self._inside_quoteblock and (
                not self._current_line or self._current_line[0] != '>'
        ):
            self._current_line += '> '
        if block.value == md4c.BlockType.P:
            self._inside_pblock = True
        elif block.value == md4c.BlockType.CODE:
            self._inside_codeblock = True
            if 'fence_char' in details:
                self._current_line += '%s' % (details['fence_char']*3)
            if details['lang']:
                self._current_line += details['lang'][0][1]
            if 'fence_char' not in details:
                self._inside_indented_codeblock = True
            if self._current_line:
                self._save_current_line()
        elif block.value == md4c.BlockType.H:
            self._current_line += '%s ' % ('#' * details['level'])
        elif block.value == md4c.BlockType.LI:
            if self._current_ol_delimiters:
                # inside OL
                if len(self._current_ol_delimiters) > 1:
                    self._save_current_msgid()
                    if not self._current_ol_delimiters[-1][0]:
                        self._save_current_line()
                self._current_ol_delimiters[-1][0] += 1
                self._current_line += '%s%d%s ' % (
                    '   ' * (len(self._current_ol_delimiters) - 1),
                    self._current_ol_delimiters[-1][0],
                    self._current_ol_delimiters[-1][1],
                )
            else:
                # inside UL
                self._current_line += '{}{} '.format(
                    '  ' * (len(self._ul_marks) - 1), self._ul_marks[-1],
                )
                if details['is_task']:
                    self._current_line += '[%s] ' % details['task_mark']
            self._inside_liblock = True
        elif block.value == md4c.BlockType.UL:
            if len(self._ul_marks) > 0:
                self._save_current_msgid()
                self._save_current_line()
            self._ul_marks.append(details['mark'])
        elif block.value == md4c.BlockType.OL:
            self._current_ol_delimiters.append([0, details['mark_delimiter']])
        elif block.value == md4c.BlockType.HR:
            self._current_line += '---'
            if not self._inside_liblock:
                self._current_line += '\n\n'
        elif block.value == md4c.BlockType.TH:
            self._current_line += '| '
            self._current_thead_aligns.append(details['align'].value)
        elif block.value == md4c.BlockType.TD:
            self._current_line += '| '
        elif block.value == md4c.BlockType.QUOTE:
            self._inside_quoteblock = True
        elif block.value == md4c.BlockType.HTML:
            self._inside_htmlblock = True

    def leave_block(self, block, details):
        # print('LEAVE BLOCK', block.name, details)

        if block.value == md4c.BlockType.P:
            self._save_current_msgid()
            if not self._inside_liblock:
                self._save_current_line()
            self._inside_pblock = False
        elif block.value == md4c.BlockType.CODE:
            self._save_current_msgid()
            self._inside_codeblock = False
            self._inside_indented_codeblock = False
            if 'fence_char' in details:
                self._current_line += '%s' % (details['fence_char']*3)
            self._save_current_line()
            self._save_current_line()
        elif block.value == md4c.BlockType.H:
            self._save_current_msgid()
            if not self._inside_quoteblock:
                self._current_line += '\n'
            else:
                self._current_line += '\n> '
            self._save_current_line()
            if self._inside_quoteblock:
                self._current_line += '> '
        elif block.value == md4c.BlockType.LI:
            self._save_current_msgid()
            self._inside_liblock = False
            self._save_current_line()
        elif block.value == md4c.BlockType.UL:
            self._ul_marks.pop()
            if self._inside_quoteblock:
                self._current_line += '> '
            if not self._ul_marks:
                self._save_current_line()
        elif block.value == md4c.BlockType.OL:
            self._current_ol_delimiters.pop()
            if self._inside_quoteblock:
                self._current_line += '> '
            if not self._current_ol_delimiters:
                self._save_current_line()
        elif block.value in (md4c.BlockType.TH, md4c.BlockType.TD):
            self._save_current_msgid()
            self._current_line += ' '
        elif block.value == md4c.BlockType.TR:
            if not self._current_thead_aligns:
                self._current_line += '|'
                self._save_current_line()
        elif block.value == md4c.BlockType.THEAD:
            self._current_line += '|'

            # build thead separator
            thead_separator = '| '
            _thead_split = re.split(r'[^\\](\|)', self._current_line)
            for i, title in enumerate(_thead_split):
                if (i % 2) != 0 or i > len(_thead_split) - 2:
                    continue
                title_length = len(title)-3
                if i == 0:
                    title_length -= 1
                title_length = max(3, title_length)
                align = self._current_thead_aligns.pop(0)
                if align in [0, 3]:
                    thead_separator += '-'
                else:
                    thead_separator += ':'
                thead_separator += ('-' * title_length)

                if align in [0, 1]:
                    thead_separator += '-'
                else:
                    thead_separator += ':'
                thead_separator += ' |'
                if i < len(_thead_split) - 3:
                    thead_separator += ' '

            self._current_line += '\n%s' % thead_separator
            self._save_current_line()
        elif block.value == md4c.BlockType.QUOTE:
            self._current_line += '> '
        elif block.value == md4c.BlockType.TABLE:
            self._save_current_line()
        elif block.value == md4c.BlockType.DOC:
            pass
        elif block.value == md4c.BlockType.HTML:
            self._inside_htmlblock = False

    def enter_span(self, span, details):
        # print("ENTER SPAN", span.name, details)

        try:
            self._current_msgid += self._enterspan_replacer[span.value]
        except KeyError:
            pass

        if span.value == md4c.SpanType.A:
            self._current_aspan_href = details['href'][0][1]
        elif span.value == md4c.SpanType.CODE:
            self._inside_codespan = True
            self._codespan_start_index = len(self._current_msgid)-1
            self._codespan_inside_current_msgid = True
        elif span.value == md4c.SpanType.IMG:
            self._current_imgspan['title'] = '' if not details['title'] \
                else details['title'][0][1]
            self._current_imgspan['src'] = details['src'][0][1]
            self._current_imgspan['text'] = ''
        elif span.value == md4c.SpanType.WIKILINK:
            self._current_wikilink_target = details['target'][0][1]

    def leave_span(self, span, details):
        # print("LEAVE SPAN", span.name, details)

        if span.value == md4c.SpanType.WIKILINK:
            self._current_msgid += polib.escape(self._current_wikilink_target)
            self._current_wikilink_target = None

        try:
            self._current_msgid += self._leavespan_replacer[span.value]
        except KeyError:
            pass

        if span.value == md4c.SpanType.A:
            if self._current_aspan_href:
                # is not self-referenced link
                self._current_msgid += '(%s' % self._current_aspan_href

                if details['title']:
                    self._aimg_title_inside_current_msgid = True
                    self._current_msgid += ' "%s"' % polib.escape(
                        details['title'][0][1],
                    )
                self._current_msgid += ')'
                self._current_aspan_href = None
        elif span.value == md4c.SpanType.CODE:
            self._inside_codespan = False
            self._current_msgid += (
                self._codespan_backticks * self.code_end_string
            )
            self._codespan_backticks = None
        elif span.value == md4c.SpanType.IMG:
            self._current_msgid += polib.escape(
                '![{}]({}'.format(
                    self._current_imgspan['text'],
                    self._current_imgspan['src'],
                ),
            )
            if self._current_imgspan['title']:
                self._current_msgid += ' "%s"' % (
                    polib.escape(self._current_imgspan['title'])
                )
            self._current_msgid += ')'
            self._current_imgspan = {}

    def text(self, block, text):
        # print("TEXT", "'%s'" % text)

        if not self._inside_htmlblock:
            if not self._inside_codeblock:
                if self._inside_liblock and text == '\n':
                    text = ' '
                if self._current_imgspan:
                    self._current_imgspan['text'] = text
                    return
                elif self._inside_codespan:
                    self._codespan_backticks = min_not_max_chars_in_a_row(
                        self.code_start_string,
                        text,
                    ) - 1
                    self._current_msgid = '{}{}{}'.format(
                        self._current_msgid[:self._codespan_start_index],
                        self._codespan_backticks * self.code_start_string,
                        self._current_msgid[self._codespan_start_index:],
                    )
                elif text == self.italic_start_string:
                    text = self.italic_start_string_escaped
                elif text == self.code_start_string:
                    text = self.code_start_string_escaped
                elif text == self.code_end_string:  # pragma: no cover
                    text = self.code_end_string_escaped
                elif text == self.italic_end_string:  # pragma: no cover
                    text = self.italic_end_string_escaped

                if self._inside_pblock:
                    text = text.replace('\n', ' ')
                if self._current_aspan_href:
                    if text == self._current_aspan_href:
                        # self-referenced link
                        self._current_aspan_href = None
                elif self._current_wikilink_target:
                    if text != self._current_wikilink_target:
                        self._current_wikilink_target = '{}|{}'.format(
                            self._current_wikilink_target, text,
                        )
                    return
                self._current_msgid += polib.escape(text)
            else:
                self._current_msgid += text
        else:
            self._process_command(text)

    def translate(self, filepath_or_content, save=None, md_encoding='utf-8'):
        content = to_file_content_if_is_file(
            filepath_or_content,
            encoding=md_encoding,
        )

        self.translations = {}
        self.translations_with_msgctxt = {}
        for pofile in self.pofiles:
            for entry in pofile:
                if entry.msgctxt:
                    if entry.msgctxt not in self.translations_with_msgctxt:
                        self.translations_with_msgctxt[entry.msgctxt] = {}
                    self.translations_with_msgctxt[
                        entry.msgctxt
                    ][entry.msgid] = entry.msgstr
                else:
                    self.translations[entry.msgid] = entry.msgstr

        parser = md4c.GenericParser(
            0,
            **{ext: True for ext in self.extensions},
        )
        parser.parse(
            content,
            self.enter_block,
            self.leave_block,
            self.enter_span,
            self.leave_span,
            self.text,
        )
        self._disable_next_line = False
        self._disable = False
        self._enable_next_line = False

        self.output = '\n'.join(self._outputlines)

        if save:
            with open(save, 'w', encoding=md_encoding) as f:
                f.write(self.output)

        return self.output


def pofile_to_markdown(
    filepath_or_content, pofiles, ignore=[],
    save=None, md_encoding='utf-8', po_encoding=None, **kwargs,
):
    r"""Translate Markdown content or a file using pofiles for message replacing.

    Args:
        filepath_or_content (str): Markdown filepath or content to translate.
        pofiles (str): Glob matching a set of pofiles from where to extract
            references to make the replacements translating strings.
        ignore (list): Paths of pofiles to ignore. Useful when a glob does not
            fit your requirements indicating the files to extract content.
            Also, filename or a dirname can be defined without indicate the
            full path.
        save (str): Saves the output content in file whose path is specified
            at this parameter.
        md_encoding (str): Markdown content encoding.
        po_encoding (str): PO files encoding. If you need different encodings
            for each file, you must define it in the "Content-Type" field of
            each PO file metadata, in the form
            ``"Content-Type: text/plain; charset=<ENCODING>\n"``.

    Returns:
        str: Markdown output file with translated content.
    """
    return Po2Md(
        pofiles, ignore=ignore, po_encoding=po_encoding,
        **kwargs,
    ).translate(
        filepath_or_content, save=save, md_encoding=md_encoding,
    )



"""Tests for mdpo markdown utilities."""

import pytest

from mdpo.md import inline_untexted_links


@pytest.mark.parametrize(
    (
        'text', 'expected_result',
    ),
    (
        ('[https://ever.com]', '<https://ever.com>'),
        ('[Ever](https://ever.com)', '[Ever](https://ever.com)'),
    ),
)
def test_inline_untexted_links__default_parameters(text, expected_result):
    assert inline_untexted_links(text) == expected_result



"""mdpo command line interface utilities."""

import argparse

from mdpo import __version__


def parse_list_cli_argument(value, splitter=','):
    """Converts values in a string separated by characters into a tuple.

    This function is needed by mdpo command line interfaces to convert
    some arguments values separated by commas into iterables.

    Args:
        value (str): String to be converted to list separating it by
            ``splitter`` argument value.
        splitter (str): Separator used for separate the ``value`` argument.

    Returns:
        tuple: Strings separated.
    """
    return tuple(filter(None, value.split(splitter)))


def add_common_cli_arguments(parser):
    """Adds common mdpo arguments to an argument parser.

    Args:
        parser (:py:class:`argparse.ArgumentParser`): Arguments parser to
            extend.
    """
    parser.add_argument(
        '-h', '--help', action='help',
        help='Show this help message and exit.',
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        '-v', '--version', action='version',
        version=f'%(prog)s {__version__}',
        help='Show program version number and exit.',
    )
    parser.add_argument(
        '-q', '--quiet', action='store_true',
        help='Do not print output to STDOUT.',
    )



"""mdpo I/O utilities."""

import glob
import os
import re


def filter_paths(filepaths, ignore_paths=[]):
    """Filters a list of paths removing those defined in other list of paths.

    The paths to filter can be defined in the list of paths to ignore in
    several forms:

    - The same string.
    - Only the file name.
    - Only their direct directory name.
    - Their direct directory full path.

    Args:
        filepaths (list): Set of source paths to filter.
        ignore_paths (list): Paths that must not be included in the response.

    Returns:
        list: Non filtered paths ordered alphabetically.
    """
    response = []
    for filepath in filepaths:
        # ignore by filename
        if os.path.basename(filepath) in ignore_paths:
            continue
        # ignore by dirname
        if os.path.basename(os.path.dirname(filepath)) in ignore_paths:
            continue
        # ignore by filepath
        if filepath in ignore_paths:
            continue
        # ignore by dirpath (relative or absolute)
        if (os.sep).join(filepath.split(os.sep)[:-1]) in ignore_paths:
            continue
        response.append(filepath)
    response.sort()
    return response


def to_file_content_if_is_file(value, encoding='utf-8'):
    """Check if the value passed is a file path or string content.

    If is a file, reads its content and returns it, otherwise returns
    the string passed as is.

    Args:
        value (str): Value to check if is a filepath or content.
        encoding (str): Expected file encoding, if is a file.

    Returns:
        str: File content if ``value`` is an existing file or ``value`` as is.
    """
    if os.path.isfile(value):
        with open(value, encoding=encoding) as f:
            value = f.read()
    return value


def to_glob_or_content(value):
    """Check if the value passed is a glob or is string content.

    Args:
        value (str): Value to check if is a glob or content.

    Returns:
        list: Two values being the first a boolean that indicates if ``value``
        is a glob (``True``) or content (``False``) and the second value
        is the content (parsed as glob is first value is ``True``).
    """
    try:
        parsed = glob.glob(value)
    except re.error:
        # some strings like '[s-m]' will produce
        # 're.error: bad character range ... at position'
        return (False, value)
    if not parsed:
        # assumes it is content
        return (False, value)
    return (True, parsed)



"""HTML-produced-from-Markdown files translator using pofiles as reference."""

import glob
import html
import re
import warnings
from html.parser import HTMLParser

import md4c
import polib

from mdpo.command import parse_mdpo_html_command
from mdpo.html import get_html_attrs_tuple_attr, html_attrs_tuple_to_string
from mdpo.io import filter_paths, to_file_content_if_is_file
from mdpo.polib import *  # noqa


PROCESS_REPLACER_TAGS = [
    'p', 'li', 'h1', 'h2', 'h3',
    'h4', 'h5', 'h6', 'td', 'th',
]
ALIGNMENT_CHARS = ['\n', ' ', '\t', '\r']


class MdPo2HTML(HTMLParser):
    def __init__(
        self, pofiles, ignore=[], merge_adjacent_markups=False,
        code_tags=['code'], bold_tags=['b', 'strong'],
        italic_tags=['em', 'i'], link_tags=['a'], image_tags=['img'],
        ignore_grouper_tags=['div', 'hr'], po_encoding=None,
    ):
        self.pofiles = [
            polib.pofile(pofilepath, encoding=po_encoding) for pofilepath in
            filter_paths(glob.glob(pofiles), ignore_paths=ignore)
        ]
        self.output = ''
        self.replacer = []
        self._raw_replacement = ''
        self.context = []
        self._current_msgctxt = None

        self.translations = None
        self.translations_with_msgctxt = None

        self._disable = False
        self._disable_next_line = False
        self._enable_next_line = False
        self.disabled_entries = []

        # lazy translators mode
        self.merge_adjacent_markups = merge_adjacent_markups

        # code markup
        self.code_tags = code_tags

        # bold markup
        self.bold_tags = bold_tags

        # italic markup
        self.italic_tags = italic_tags

        # link markup
        self.link_tags = link_tags

        # image markup
        self.image_tags = image_tags

        self.ignore_grouper_tags = ignore_grouper_tags

        self.markup_tags = []
        self.markup_tags.extend(self.code_tags)
        self.markup_tags.extend(self.bold_tags)
        self.markup_tags.extend(self.italic_tags)

        self.html_renderer = md4c.HTMLRenderer(md4c.MD_FLAG_TABLES)

        super().__init__()

    def _merge_adyacent_tags(self, html, template_tags):
        for tags_group in [self.bold_tags, self.italic_tags]:
            regexes = []
            for tag in tags_group:
                if tag not in template_tags:
                    continue
                for _tag in tags_group:
                    if _tag not in template_tags:
                        continue
                    regex = fr'</{tag}>\s*<{_tag}>'
                    if regex not in regexes:
                        regexes.append(regex)

            for regex in regexes:
                html = re.sub(regex, ' ', html)

        return html

    def _remove_lastline_from_output_if_empty(self):
        split_output = self.output.split('\n')
        if not split_output[-1]:
            self.output = '\n'.join(split_output)[:-1]

    def _process_replacer(self):
        # print("REPLACER:", self.replacer)

        template_tags = []
        raw_html_template, _current_replacement = ('', '')

        _current_link_target = ''

        _last_start_tag = None
        _last_end_tag = None
        _inside_code = False
        while self.replacer:
            handle, handled, attrs = self.replacer.pop(0)
            if handle == 'start':
                template_tags.append(handled)

                if handled in self.code_tags:
                    _current_replacement += '`'
                    raw_html_template += '<{}{}>'.format(
                        handled,
                        (
                            ' ' + html_attrs_tuple_to_string(attrs)
                            if attrs else ''
                        ),
                    )
                elif handled in self.bold_tags:
                    _current_replacement += '**'
                    raw_html_template += '<{}{}>'.format(
                        handled,
                        (
                            ' ' + html_attrs_tuple_to_string(attrs)
                            if attrs else ''
                        ),
                    )
                elif handled in self.italic_tags:
                    _current_replacement += '*'
                    raw_html_template += '<{}{}>'.format(
                        handled,
                        (
                            ' ' + html_attrs_tuple_to_string(attrs)
                            if attrs else ''
                        ),
                    )
                elif handled in self.link_tags:
                    title = get_html_attrs_tuple_attr(attrs, 'title')
                    _current_link_target += '(%s' % get_html_attrs_tuple_attr(
                        attrs, 'href',
                    )
                    if title:
                        _current_link_target += ' "%s"' % title
                    _current_link_target += ')'

                    raw_html_template += '<%s' % handled

                    # attrs_except_href_title = []
                    for attr, value in attrs:
                        if attr in ['title', 'href']:
                            raw_html_template += ' %s="{}"' % attr
                        # else:
                        #     These attributes are not included in output
                        #    attrs_except_href_title.append((attr, value))
                    # if attrs_except_href_title:
                    #     raw_html_template += ' '
                    # raw_html_template += html_attrs_tuple_to_string(
                    #     attrs_except_href_title) + '>'
                    raw_html_template += '>'
                else:
                    raw_html_template += '<{}{}>'.format(
                        handled,
                        ' ' + html_attrs_tuple_to_string(attrs)
                        if attrs else '',
                    )
                _last_start_tag = handled
                if _last_start_tag == 'code':
                    _inside_code = True

            elif handle == 'data':
                if not _inside_code:
                    handled = re.sub(r'\s{2,}', ' ', handled)

                if all((ch in ALIGNMENT_CHARS) for ch in handled):
                    raw_html_template += handled
                    _current_replacement += handled
                else:
                    raw_html_template += '{}'
                    if _current_link_target:
                        _current_replacement += '[{}]{}'.format(
                            handled, _current_link_target,
                        )
                        _current_link_target = ''
                    else:
                        _current_replacement += handled
            elif handle == 'end':
                raw_html_template += '</%s>' % handled
                if handled in self.code_tags:
                    _current_replacement += '`'
                elif handled in self.bold_tags:
                    _current_replacement += '**'
                elif handled in self.italic_tags:
                    _current_replacement += '*'
                _last_end_tag = handled
                if _last_end_tag == 'code':
                    _inside_code = False
            elif handle == 'comment':
                raw_html_template += '<!--%s-->' % handled
            elif handle == 'startend':
                if handled in self.image_tags:
                    _current_replacement += '![{}]({}'.format(
                        get_html_attrs_tuple_attr(attrs, 'alt'),
                        get_html_attrs_tuple_attr(attrs, 'src'),
                    )
                    title = get_html_attrs_tuple_attr(attrs, 'title')
                    if title:
                        _current_replacement += ' "%s"' % title
                    _current_replacement += ')'

                    raw_html_template += '{}'
                else:
                    raw_html_template += '<{}{}/>'.format(
                        handled,
                        (
                            (' %s' % html_attrs_tuple_to_string(attrs))
                            if attrs else ''
                        ),
                    )

        _current_replacement = html.unescape(_current_replacement)

        if (self._disable and not self._enable_next_line) \
                or self._disable_next_line:
            replacement = _current_replacement

            self.disabled_entries.append(
                polib.POEntry(
                    msgid=replacement,
                    msgstr='',
                    msgctxt=self._current_msgctxt,
                ),
            )
        else:
            if self._current_msgctxt:
                replacement = self.translations_with_msgctxt[
                    self._current_msgctxt
                ].get(_current_replacement)
            else:
                replacement = self.translations.get(_current_replacement)
            if not replacement:
                replacement = _current_replacement

        for tt in template_tags:
            for tags_group in [self.bold_tags, self.italic_tags]:
                if tt in tags_group:
                    for tag in tags_group:
                        if tag not in template_tags:
                            template_tags.append(tag)

        # print("RAW TEMPLATE:", raw_html_template)
        # print("TEMPLATE TAGS:", template_tags)
        # print('CURRENT MSGID: \'%s\'' % _current_replacement)
        # print('MSGSTR:', replacement)

        html_before_first_replacement = raw_html_template.split('{')[0]
        html_after_last_replacement = raw_html_template.split('}')[-1]
        for tags_group in [self.bold_tags, self.italic_tags, self.code_tags]:
            for tag in tags_group:
                html_before_first_replacement = \
                    html_before_first_replacement.split('<%s>' % tag)[0]
                html_after_last_replacement = \
                    html_after_last_replacement.split('</%s>' % tag)[-1]
            html_before_first_replacement = \
                html_before_first_replacement.split('<a href="')[0]

        html_inner = '\n'.join(
            self.html_renderer.parse(replacement).split('\n')[:-1],
        )
        html_inner = '>'.join(html_inner.split('>')[1:])
        html_inner = '<'.join(html_inner.split('<')[:-1])

        html_template = html_before_first_replacement + html_inner + \
            html_after_last_replacement

        if self.merge_adjacent_markups:
            html_template = self._merge_adyacent_tags(
                html_template,
                template_tags,
            )

        self.output += html_template
        self.context = []

        self._disable_next_line = False
        self._enable_next_line = False
        self._current_msgctxt = None

        # print("________________________________________________")

    def handle_starttag(self, tag, attrs):
        # print("START TAG: %s | POS: %d:%d" % (tag, *self.getpos()))

        if tag in self.ignore_grouper_tags:
            self.context.append(tag)
            self.output += '<{}{}>'.format(
                tag, ' ' + html_attrs_tuple_to_string(attrs) if attrs else '',
            )
        elif self.context and self.context[0] in self.ignore_grouper_tags:
            self.output += '<{}{}>'.format(
                tag, ' ' + html_attrs_tuple_to_string(attrs) if attrs else '',
            )
        elif tag == 'ul' and not self.context:
            self.output += '<{}{}>'.format(
                tag, ' ' + html_attrs_tuple_to_string(attrs) if attrs else '',
            )
        elif tag in ['blockquote', 'table', 'thead', 'tbody', 'tr']:
            self.output += '<{}{}>'.format(
                tag, ' ' + html_attrs_tuple_to_string(attrs) if attrs else '',
            )
        else:
            self.replacer.append(('start', tag, attrs))
            self.context.append(tag)

    def handle_endtag(self, tag):
        # print("END TAG: %s | POS: %d:%d" % (tag, *self.getpos()))

        if tag in self.ignore_grouper_tags:
            self.output += '</%s>' % tag
            if self.context:
                self.context.pop()
        elif self.context and self.context[0] in self.ignore_grouper_tags:
            self.output += '</%s>' % tag
        elif tag in PROCESS_REPLACER_TAGS:
            self.replacer.append(('end', tag, None))
            self._process_replacer()
        elif tag in ['ul', 'blockquote', 'tr', 'table', 'thead', 'tbody']:
            self.output += '</%s>' % tag
        else:
            self.replacer.append(('end', tag, None))
            if self.context:
                self.context.pop()

    def handle_startendtag(self, tag, attrs):
        # print("STARTEND TAG: %s | POS: %d:%d" % (tag, *self.getpos()))

        if not self.replacer:
            self.output += self.get_starttag_text()
        else:
            self.replacer.append(('startend', tag, attrs))

    def handle_data(self, data):
        # print("     DATA: '%s'" % (data))

        if data:
            if not self.replacer or (
                    self.context and
                    self.context[0] in self.ignore_grouper_tags
            ) or \
                    not self.context:
                self.output += data
                return
            else:
                if self.context:
                    data = data.replace('\n', ' ')
                self.replacer.append(('data', data, None))

    def handle_comment(self, data):
        # print("     COMMENT: '%s'" % (data))

        if self.replacer:
            self.replacer.append(('comment', data, None))
        else:
            data_as_comment = '<!--%s-->' % data
            command, comment = parse_mdpo_html_command(data_as_comment)
            if command is None:
                self.output += data_as_comment
            else:
                self._remove_lastline_from_output_if_empty()
                if command == 'disable-next-line':
                    self._disable_next_line = True
                elif command == 'disable':
                    self._disable = True
                elif command == 'enable':
                    self._disable = False
                elif command == 'enable-next-line':
                    self._enable_next_line = True
                elif command == 'context' and comment:
                    self._current_msgctxt = comment.strip(' ')
                elif command == 'include-codeblock':
                    warnings.warn(
                        'Code blocks translations is not supported'
                        ' by mdpo2html implementation.',
                        SyntaxWarning,
                    )

    def translate(self, filepath_or_content, save=None, html_encoding='utf-8'):
        content = to_file_content_if_is_file(
            filepath_or_content,
            encoding=html_encoding,
        )

        self.translations, self.translations_with_msgctxt = ({}, {})
        for pofile in self.pofiles:
            for entry in pofile:
                if entry.msgctxt:
                    if entry.msgctxt not in self.translations_with_msgctxt:
                        self.translations_with_msgctxt[entry.msgctxt] = {}
                    self.translations_with_msgctxt[
                        entry.msgctxt
                    ][entry.msgid] = entry.msgstr
                else:
                    self.translations[entry.msgid] = entry.msgstr

        self.feed(content)

        if save:
            with open(save, 'w', encoding=html_encoding) as f:
                f.write(self.output)

        self.reset()

        return self.output


def markdown_pofile_to_html(
    filepath_or_content, pofiles, ignore=[], save=None,
    po_encoding=None, html_encoding='utf-8', **kwargs,
):
    r"""Produces a translated HTML file given a previous HTML file (created by a
    Markdown-to-HTML processor) and a set of pofiles as reference for msgstrs.

    Args:
        filepath_or_content (str): HTML whose content wants to be translated.
        pofiles (str): Glob for set of pofiles used as reference translating
            the strings to another language.
        ignore (list): List of paths to pofiles to ignore, useful if the glob
            patterns in ``pofiles`` parameter does not fit your requirements.
        save (str): If you pass this parameter as a path to one HTML file,
            even if does not exists, will be saved in the path the output of
            the function.
        md_encoding (str): HTML content encoding.
        po_encoding (str): PO files encoding. If you need different encodings
            for each file, you must define it in the "Content-Type" field of
            each PO file metadata, in the form
            ``"Content-Type: text/plain; charset=<ENCODING>\n"``.

    .. rubric:: Known limitations:

    * This implementation doesn't include support for
      :ref:`include-codeblock command<include-codeblock-command>`.


    Returns:
        str: HTML output translated version of the given file.
    """
    return MdPo2HTML(
        pofiles, ignore=ignore, po_encoding=po_encoding, **kwargs,
    ).translate(filepath_or_content, save=save, html_encoding=html_encoding)



"""md4c related stuff for mdpo."""


#: :list: `md4c parser <https://github.com/mity/md4c>`_ extensions used as
#: default by :ref:`md2po<md2po-init>` and :ref:`po2md<po2md-init>`
#: implementations.
DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS = [
    'collapse_whitespace',
    'tables',
    'strikethrough',
    'tasklists',
    'latex_math_spans',
    'wikilinks',
]
